/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "sokoke.h"

void sokoke_combo_box_add_strings(GtkComboBox* combobox
 , const gchar* sLabelFirst, ...)
{
    // Add a number of strings to a combobox, terminated with NULL
    // This works only for text comboboxes
    va_list args;
    va_start(args, sLabelFirst);

    const gchar* sLabel;
    for(sLabel = sLabelFirst; sLabel; sLabel = va_arg(args, const gchar*))
     gtk_combo_box_append_text(combobox, sLabel);

    va_end(args);
}

void sokoke_combo_box_add_strings_with_ints(GtkComboBox* combobox
 , const gchar* sLabelFirst, gint iValueFirst, ...)
{
    // Add a number of strings and ints to a text combo box, terminated with NULL
    // FIXME: We assume blindly that 0 is the string and 1 is the int
    va_list args;
    va_start(args, iValueFirst);

    /*const gchar* sLabel; gint iValue;
    for(sLabel = sLabelFirst, iValue = iValueFirst; sLabel
     ; sLabel = va_arg(args, const gchar*), iValue = va_arg(args, gint))
    {
     GtkTreeIter iter;
     gtk_list_store_append(store, &iter);
     gtk_list_store_set(store, &iter , 0, sLabel, 1, iValue, -1);
    }*/

    va_end(args);
}

void sokoke_radio_action_set_current_value(GtkRadioAction* action
 , gint current_value)
{
    // Activates the group member with the given value
    #if GTK_CHECK_VERSION(2, 10, 0)
    gtk_radio_action_set_current_value(action, current_value);
    #else
    // TODO: Implement this for older gtk
    #endif
}

gboolean sokoke_de_is_xfce(void)
{
    // Are we running in Xfce?
    gint iResult; gchar* stdout; gchar* stderr;
    gboolean bSuccess = g_spawn_command_line_sync(
     "xprop -root _DT_SAVE_MODE | grep -q xfce4"
     , &stdout, &stderr, &iResult, NULL);
    return bSuccess && !iResult;
}

void sokoke_widget_set_visible(GtkWidget* widget, gboolean bVisibility)
{
    // Show or hide the widget
    if(bVisibility) gtk_widget_show(widget); else gtk_widget_hide(widget);
}

void sokoke_container_show_children(GtkContainer* container)
{
    // Show every child but not the container itself
    gtk_container_foreach(container, (GtkCallback)(gtk_widget_show_all), NULL);
}

GtkWidget* sokoke_hig_frame_new(const gchar* sLabel)
{
    // Create a frame with no actual frame but a bold label and indentation
    GtkWidget* frame = gtk_frame_new(NULL);
    gchar* sLabelBold = g_strdup_printf("<b>%s</b>", sLabel);
    GtkWidget* label = gtk_label_new(NULL);
    gtk_label_set_markup(GTK_LABEL(label), sLabelBold);
    g_free(sLabelBold);
    gtk_frame_set_label_widget(GTK_FRAME(frame), label);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);
    return frame;
}

void sokoke_widget_set_pango_font_style(GtkWidget* widget, PangoStyle style)
{
    // Conveniently change the pango font style
    // For some reason we need to reset if we actually want the normal style
    if(style == PANGO_STYLE_NORMAL)
     gtk_widget_modify_font(widget, NULL);
    else
    {
     PangoFontDescription* pangofontdesc = pango_font_description_new();
     pango_font_description_set_style(pangofontdesc, PANGO_STYLE_ITALIC);
     gtk_widget_modify_font(widget, pangofontdesc);
     pango_font_description_free(pangofontdesc);
    }
}

gboolean sokoke_on_entry_focus_in_event(GtkEntry* entry, GdkEventFocus *event
 , gpointer userdata)
{
    gboolean bDefaultText = (gboolean)g_object_get_data(G_OBJECT(entry)
     , "sokoke_bDefaultText");
    if(bDefaultText)
    {
     gtk_entry_set_text(entry, "");
     g_object_set_data(G_OBJECT(entry), "sokoke_bDefaultText", (gpointer)FALSE);
     sokoke_widget_set_pango_font_style(GTK_WIDGET(entry), PANGO_STYLE_NORMAL);
    }
    return FALSE;
}

gboolean sokoke_on_entry_focus_out_event(GtkEntry* entry, GdkEventFocus *event
 , gpointer userdata)
{
    const gchar* sText = gtk_entry_get_text(entry);
    if(!strlen(sText))
    {
     const gchar* sDefaultText = (const gchar*)g_object_get_data(
      G_OBJECT(entry), "sokoke_sDefaultText");
     gtk_entry_set_text(entry, sDefaultText);
     g_object_set_data(G_OBJECT(entry), "sokoke_bDefaultText", (gpointer)TRUE);
     sokoke_widget_set_pango_font_style(GTK_WIDGET(entry), PANGO_STYLE_ITALIC);
    }
    return FALSE;
}

void sokoke_entry_set_default_text(GtkEntry* entry, const gchar* sDefaultText)
{
    // FIXME: Calling this on an entry which has text is undefined
    g_object_set_data(G_OBJECT(entry), "sokoke_sDefaultText", (gpointer)sDefaultText);
    g_object_set_data(G_OBJECT(entry), "sokoke_bDefaultText", (gpointer)TRUE);
    gtk_entry_set_text(entry, sDefaultText);
    sokoke_widget_set_pango_font_style(GTK_WIDGET(entry), PANGO_STYLE_ITALIC);
    g_signal_connect(G_OBJECT(entry), "focus-in-event"
     , G_CALLBACK(sokoke_on_entry_focus_in_event), NULL);
    g_signal_connect(G_OBJECT(entry), "focus-out-event"
     , G_CALLBACK(sokoke_on_entry_focus_out_event), NULL);
}

gchar* sokoke_key_file_get_string_default(GKeyFile* key_file
 , const gchar* group_name, const gchar* key, const gchar* def, GError* *error)
{
    gchar* value = g_key_file_get_string(key_file, group_name, key, error);
    return value == NULL ? g_strdup(def) : value;
}

gint sokoke_key_file_get_integer_default(GKeyFile* key_file
 , const gchar* group_name, const gchar* key, const gint def, GError* *error)
{
    if(!g_key_file_has_key(key_file, group_name, key, NULL)) return def;
    return g_key_file_get_integer(key_file, group_name, key, error);
}

gboolean sokoke_key_file_save_to_file(GKeyFile* key_file
 , const gchar* file, GError* *error)
{
    gchar* data = g_key_file_to_data(key_file, NULL, error);
    FILE *fp;
    if((fp = fopen(file, "w")) == NULL) return FALSE;
    fprintf(fp, data);
    fclose(fp);
    g_free(data);
    return TRUE;
}

void sokoke_widget_get_text_size(GtkWidget* widget, const gchar* sText
 , gint* w, gint* h)
{
    PangoLayout* layout = gtk_widget_create_pango_layout(widget, sText);
    pango_layout_get_pixel_size(layout, w, h);
    g_object_unref(layout);
}

#define QUIT_LOOP(ri) \
 if(g_main_loop_is_running(ri->loop)) \
  g_main_loop_quit (ri->loop)

static void sokoke_unmap_h(GtkDialog* dialog, sokoke_ri* ri)
{
    QUIT_LOOP(ri);
}

static void sokoke_response_h(GtkDialog*dialog, gint response_id, sokoke_ri* ri)
{
    ri->response_id = response_id;
    QUIT_LOOP(ri);
}

static gint sokoke_delete_h(GtkDialog* dialog, GdkEventAny* event, sokoke_ri* ri)
{
    QUIT_LOOP(ri);
    return TRUE;
}

static void sokoke_destroy_h(GtkDialog* dialog, sokoke_ri* ri)
{
    // QUIT_LOOP will be called in sokoke_unmap_h
    ri->destroyed = TRUE;
}

gint sokoke_dialog_run_modeless(GtkDialog* dialog)
{
    // Based on gtk_dialog_run from gtk/gtkdialog.c from the GTK+ project.
    // Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald

    // This will display a modeless dialog and return the response

    sokoke_ri ri = { NULL, GTK_RESPONSE_NONE, NULL, FALSE };

    g_return_val_if_fail(GTK_IS_DIALOG(dialog), -1);
    g_object_ref(dialog);

    if(!GTK_WIDGET_VISIBLE(dialog)) gtk_widget_show(GTK_WIDGET(dialog));

    gulong response_h = g_signal_connect(dialog, "response"
     , GTK_SIGNAL_FUNC(sokoke_response_h), &ri);
    gulong unmap_h = g_signal_connect(dialog, "unmap"
     , GTK_SIGNAL_FUNC(sokoke_unmap_h), &ri);
    gulong delete_h = g_signal_connect(dialog, "delete-event"
     , GTK_SIGNAL_FUNC(sokoke_destroy_h), &ri);
    gulong destroy_h = g_signal_connect(dialog, "destroy"
     , GTK_SIGNAL_FUNC(sokoke_destroy_h), &ri);

    ri.loop = g_main_loop_new (NULL, FALSE);
    GDK_THREADS_LEAVE();
    g_main_loop_run(ri.loop);
    GDK_THREADS_ENTER();
    g_main_loop_unref(ri.loop);
    ri.loop = NULL;

    if(!ri.destroyed)
    {
      g_signal_handler_disconnect(dialog, response_h);
      g_signal_handler_disconnect(dialog, unmap_h);
      g_signal_handler_disconnect(dialog, delete_h);
      g_signal_handler_disconnect(dialog, destroy_h);
    }

    g_object_unref (dialog);
    return ri.response_id;
}

#undef QUIT_LOOP

void sokoke_widget_update_screen(GtkWidget* widget
 , GdkScreen* old_screen, gpointer userdata)
{
    GdkScreen* screen = gtk_widget_get_screen(widget);
    GdkColormap* colormap;
    #if GTK_CHECK_VERSION(2, 8, 0)
    colormap = gdk_screen_get_rgba_colormap(screen);
    if(!colormap)
    {
    #endif
     colormap = gdk_screen_get_rgb_colormap(screen);
    #if GTK_CHECK_VERSION(2, 8, 0)
    }
    #endif
    gtk_widget_set_colormap(widget, colormap);
}

static gboolean sokoke_on_transparent_widget_expose(GtkWidget* widget
 , GdkEventExpose* expose, gpointer userdata)
{
    // Based on on_alpha_window_expose from gtk/tests/testgtk.c from the GTK+ project.
    // Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald

    // This will render the given widget partially transparent
    // TODO: This is GTK+ 2.8 or GTK+ 2.10 only
    // TODO: We don't want that circular stuff eventually
    // TODO: This doesn't really work on anything but windows
    // TODO: This fails on widgets which paint a background
    #if GTK_CHECK_VERSION(2, 10, 0) //...
    cairo_t* cr;
    cairo_pattern_t* pattern;
    gint radius;
    gint iAlpha = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(widget), "sokoke_iAlpha"));

    cr = gdk_cairo_create(widget->window);
    radius = MIN(widget->allocation.width, widget->allocation.height) / 2;
    pattern = cairo_pattern_create_radial(widget->allocation.width / 2
     , widget->allocation.height / 2, 0.0
     , widget->allocation.width / 2, widget->allocation.height / 2
     , radius * 1.33);

    if(gdk_screen_get_rgba_colormap(gtk_widget_get_screen(widget))
     && gtk_widget_is_composited(widget))
     cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, iAlpha / 10);
    else
     cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);

    cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
    cairo_paint(cr);
    cairo_pattern_add_color_stop_rgba(pattern, 0.0, 1.0, 0.75, 0.0, 1.0);
    cairo_pattern_add_color_stop_rgba(pattern, 1.0, 1.0, 0.75, 0.0, 0.0);
    cairo_set_source(cr, pattern);
    cairo_pattern_destroy(pattern);
    cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
    cairo_paint(cr);
    cairo_destroy(cr);
    #endif

    return FALSE;
}

void sokoke_widget_set_transparency(GtkWidget* widget, gfloat fAlpha)
{
    // FIXME: Only one decimal is allowed, any more are stripped
    gint iAlpha = fAlpha * 10;
    gtk_widget_set_app_paintable(widget, iAlpha > 0);
    g_object_set_data(G_OBJECT(widget), "sokoke_iAlpha", GINT_TO_POINTER(iAlpha));
    if(fAlpha)
    {
     g_signal_connect(G_OBJECT(widget), "expose-event"
      , G_CALLBACK(sokoke_on_transparent_widget_expose), NULL);
     g_signal_connect(G_OBJECT(widget), "screen-changed"
      , G_CALLBACK(sokoke_widget_update_screen), NULL);
     sokoke_widget_update_screen(widget, NULL, NULL);
    }
    else
    {
     // TODO: disconnect signals
    }
}

void sokoke_menu_item_set_accel(GtkMenuItem* menuitem, const gchar* sPath
 , const gchar* sKey, GdkModifierType accel_mods)
{
    if(strlen(sPath))
    {
     gchar* path = g_strconcat("<", g_get_prgname(), ">/", sPath, NULL);
     gtk_menu_item_set_accel_path(GTK_MENU_ITEM(menuitem), path);
     guint keyVal = sKey ? gdk_keyval_from_name(sKey) : 0;
     gtk_accel_map_add_entry(path, keyVal, accel_mods);
     g_free(path);
    }
}

gboolean sokoke_entry_can_undo(GtkEntry* entry)
{
    // TODO: Can we undo the last input?
    return FALSE;
}

gboolean sokoke_entry_can_redo(GtkEntry* entry)
{
    // TODO: Can we redo the last input?
    return FALSE;
}

void sokoke_entry_undo(GtkEntry* entry)
{
    // TODO: Implement undo
}

void sokoke_entry_redo(GtkEntry* entry)
{
    // TODO: Implement redo
}

gboolean sokoke_on_undo_entry_key_down(GtkEntry* widget, GdkEventKey* event
 , gpointer userdata)
{
    switch(event->keyval)
    {
     case GDK_Undo:
      sokoke_entry_undo(widget);
      return FALSE;
     case GDK_Redo:
      sokoke_entry_redo(widget);
      return FALSE;
     default:
      return FALSE;
    }
}

void sokoke_on_undo_entry_populate_popup(GtkEntry* entry, GtkMenu* menu
 , gpointer userdata)
{
    // Enhance the entry's menu with undo and redo items.
    GtkWidget* menuitem = gtk_separator_menu_item_new();
    gtk_menu_shell_prepend((GtkMenuShell*)menu, menuitem);
    gtk_widget_show(menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_REDO, NULL);
    g_signal_connect(menuitem, "activate", GTK_SIGNAL_FUNC(sokoke_entry_redo), userdata);
    gtk_menu_shell_prepend((GtkMenuShell*)menu, menuitem);
    gtk_widget_show(menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_UNDO, NULL);
    g_signal_connect(menuitem, "activate", GTK_SIGNAL_FUNC(sokoke_entry_undo), userdata);
    gtk_menu_shell_prepend((GtkMenuShell*)menu, menuitem);
    gtk_widget_show(menuitem);
}

gboolean sokoke_entry_get_can_undo(GtkEntry* entry)
{
    // TODO: Is this entry undo enabled?
    return FALSE;
}

void sokoke_entry_set_can_undo(GtkEntry* entry, gboolean bCanUndo)
{
    if(bCanUndo)
    {
     g_signal_connect(G_OBJECT(entry), "key-press-event"
      , G_CALLBACK(sokoke_on_undo_entry_key_down), NULL);
     g_signal_connect(G_OBJECT(entry), "populate-popup"
      , G_CALLBACK(sokoke_on_undo_entry_populate_popup), NULL);
    }
    else
    {
     ; // TODO: disconnect signal
    }
}
